###############################################################################
# PyDial: Multi-domain Statistical Spoken Dialogue System Software
###############################################################################
#
# Copyright 2015 - 2019
# Cambridge University Engineering Department Dialogue Systems Group
#
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

'''
DQNPolicy.py - deep Q network policy
==================================================

Author: Pei-Hao (Eddy) Su  (Copyright CUED Dialogue Systems Group 2016)

.. seealso:: CUED Imports/Dependencies: 

    import :class:`Policy`
    import :class:`utils.ContextLogger`

.. warning::
        Documentation not done.


************************

'''

import copy
import sys
import os
import json
import numpy as np
import pickle as pickle
import random
import utils
from utils.Settings import config as cfg
from utils import ContextLogger

import ontology.FlatOntologyManager as FlatOnt
#from theano_dialogue.util.tool import *

import tensorflow as tf
from policy.DRL.replay_buffer import ReplayBuffer
from policy.DRL.replay_prioritised import ReplayPrioritised
import policy.DRL.utils as drlutils
import policy.DRL.dqn as dqn
import policy.Policy
import policy.SummaryAction
from policy.Policy import TerminalAction, TerminalState
import policy.GPPolicy

logger = utils.ContextLogger.getLogger('')

# --- for flattening the belief --- # 
domainUtil = FlatOnt.FlatDomainOntology('CamRestaurants')

"""
def flatten_belief(gpstate):
    '''
    Flatten the GP-dictionary-typed belief state to a one-dim vector
    '''

    if isinstance(gpstate, TerminalState):
        return [0] * 304 #260 #264

    flat_belief = []
    for key, value in gpstate._bstate.items():
        flat_belief += value

    return flat_belief
"""

def flatten_belief(belief,domainUtil=FlatOnt.FlatDomainOntology('CamRestaurants'), merge=False):
    if isinstance(belief, TerminalState):
        return [0] * 260 #264

    #for key, value in belief.items():
    #    print key, value

    #policyfeatures = ['full','method','discourseAct','requested']
    policyfeatures = ['full','method','discourseAct','requested',\
                'lastActionInformNone','offerHappened','inform_info']

    flat_belief = []
    for feat in policyfeatures:
        add_feature = []
        if feat == 'kbest':
            for slot in self.domainUtil.sorted_system_requestable_slots:
                # print slot, 'belief', belief['beliefs'][slot]
                temp = [belief['beliefs'][slot][value] for value in domainUtil.ontology['informable'][slot]]
                temp = sorted(temp, key=lambda b: -b)
               #temp = [belief['beliefs'][slot]['dontcare']] + [belief['beliefs'][slot]['**NONE**']] + temp
                temp = temp + [belief['beliefs'][slot]['dontcare']] + [belief['beliefs'][slot]['**NONE**']]
                temp = temp[0:self.max_k]
                add_feature += temp
        elif feat == 'full':
            #for slot in self.sorted_slots:
            for slot in domainUtil.ontology['informable']:
                for value in domainUtil.ontology['informable'][slot]:# + ['**NONE**']:
                #for value in domainUtil.ontology['informable'][slot] + ['**NONE**']:
                #for value in domainUtil.ontology['informable'][slot] + ['dontcare'] + ['**NONE**']:
                    add_feature.append(belief['beliefs'][slot][value])
        elif feat == 'method':
            add_feature = [belief['beliefs']['method'][method] for method in domainUtil.ontology['method']]
        elif feat == 'discourseAct':
            add_feature = [belief['beliefs']['discourseAct'][discourseAct]
                           for discourseAct in domainUtil.ontology['discourseAct']]
        elif feat == 'requested':
            add_feature = [belief['beliefs']['requested'][slot] \
                            for slot in domainUtil.ontology['requestable']]
        elif feat == 'lastActionInformNone':
            add_feature.append(float(belief['features']['lastActionInformNone']))
        elif feat == 'offerHappened':
            add_feature.append(float(belief['features']['offerHappened']))
        elif feat == 'inform_info':
            add_feature += belief['features']['inform_info']
        else:
            logger.error('Invalid feature name in config: ' + feat)

        flat_belief += add_feature

    return flat_belief

    
    
    """ 
    flat_belief = []
    for feat in policyfeatures:
        add_feature = []
        if feat == 'full':
            #for slot in self.sorted_slots:
            for slot in domainUtil.ontology['informable']:
                if slot == 'name':
                    continue
                accumProb = 0.0
                for value in domainUtil.ontology['informable'][slot] + ['**NONE**']:
                     if value not in ('dontcare', '**NONE**'):
                        accumProb += float(belief['beliefs'][slot][value])
                add_feature.append(accumProb)
                add_feature.append(belief['beliefs'][slot]['dontcare'])
                add_feature.append(belief['beliefs'][slot]['**NONE**'])

                   #add_feature.append(belief['beliefs'][slot][value])
        elif feat == 'method':
            add_feature = [belief['beliefs']['method'][method] \
                            for method in domainUtil.ontology['method']]
        elif feat == 'discourseAct':
            add_feature = [belief['beliefs']['discourseAct'][discourseAct]
                           for discourseAct in domainUtil.ontology['discourseAct']]
        elif feat == 'requested':
            add_feature = [belief['beliefs']['requested'][slot] \
                            for slot in domainUtil.ontology['requestable']]
        elif feat == 'lastActionInformNone':
            add_feature.append(float(belief['features']['lastActionInformNone']))
        elif feat == 'offerHappened':
            add_feature.append(float(belief['features']['offerHappened']))
        elif feat == 'inform_info':
            add_feature += (belief['features']['inform_info'])
        else:
            logger.error('Invalid feature name in config: ' + feat)

        flat_belief += add_feature
    return flat_belief
    """ 


class DQNPolicy(Policy.Policy):
    '''Derived from :class:`Policy`
    '''
    def __init__(self, in_policy_file, out_policy_file, domainString='CamRestaurants', is_training=False):
        super(DQNPolicy, self).__init__(domainString, is_training)

        self.in_policy_file = in_policy_file
        self.out_policy_file = out_policy_file
        self.is_training = is_training
        self.accum_belief = []
        self.stats = [0 for ii in range(14)]

        self.prev_state_check = None

        # parameter settings
        self.n_in= 260
        if cfg.has_option('dqnpolicy_'+domainString, 'n_in'):
            self.n_in = cfg.getint('dqnpolicy_'+domainString, 'n_in')

        self.actor_lr = 0.0001
        if cfg.has_option('dqnpolicy_'+domainString, 'actor_lr'):
            self.actor_lr = cfg.getfloat('dqnpolicy_'+domainString, 'actor_lr')

        self.critic_lr = 0.001
        if cfg.has_option('dqnpolicy_'+domainString, 'critic_lr'):
            self.critic_lr = cfg.getfloat('dqnpolicy_'+domainString, 'critic_lr')

        self.tau = 0.001
        if cfg.has_option('dqnpolicy_'+domainString, 'tau'):
            self.tau = cfg.getfloat('dqnpolicy_'+domainString, 'tau')

        self.randomseed = 1234
        if cfg.has_option('GENERAL', 'seed'):
            self.randomseed = cfg.getint('GENERAL', 'seed')
        
        self.gamma = 1.0
        if cfg.has_option('dqnpolicy_'+domainString, 'gamma'):
            self.gamma = cfg.getfloat('dqnpolicy_'+domainString, 'gamma')

        self.regularisation = 'l2'
        if cfg.has_option('dqnpolicy_'+domainString, 'regularisation'):
            self.regularisation = cfg.get('dqnpolicy_'+domainString, 'regulariser')

        self.learning_rate = 0.001
        if cfg.has_option('dqnpolicy_'+domainString, 'learning_rate'):
            self.learning_rate = cfg.getfloat('dqnpolicy_'+domainString, 'learning_rate')

        self.exploration_type = 'e-greedy' # Boltzman
        if cfg.has_option('dqnpolicy_'+domainString, 'exploration_type'):
            self.exploration_type = cfg.get('dqnpolicy_'+domainString, 'exploration_type')

        self.episodeNum = 1000
        if cfg.has_option('dqnpolicy_'+domainString, 'episodeNum'):
            self.episodeNum = cfg.getfloat('dqnpolicy_'+domainString, 'episodeNum')

        self.maxiter = 5000
        if cfg.has_option('dqnpolicy_'+domainString, 'maxiter'):
            self.maxiter = cfg.getfloat('dqnpolicy_'+domainString, 'maxiter')

        self.epsilon = 1
        if cfg.has_option('dqnpolicy_'+domainString, 'epsilon'):
            self.epsilon = cfg.getfloat('dqnpolicy_'+domainString, 'epsilon')
        
        self.epsilon_start = 1
        if cfg.has_option('dqnpolicy_'+domainString, 'epsilon_start'):
            self.epsilon_start = cfg.getfloat('dqnpolicy_'+domainString, 'epsilon_start')

        self.epsilon_end = 1
        if cfg.has_option('dqnpolicy_'+domainString, 'epsilon_end'):
            self.epsilon_end = cfg.getfloat('dqnpolicy_'+domainString, 'epsilon_end')

        self.priorProbStart = 1.0
        if cfg.has_option('dqnpolicy_'+domainString, 'prior_sample_prob_start'):
            self.priorProbStart = cfg.getfloat('dqnpolicy_'+domainString, 'prior_sample_prob_start')

        self.priorProbEnd = 0.1
        if cfg.has_option('dqnpolicy_'+domainString, 'prior_sample_prob_end'):
            self.priorProbEnd = cfg.getfloat('dqnpolicy_'+domainString, 'prior_sample_prob_end')

        self.policyfeatures = []
        if cfg.has_option('dqnpolicy_'+domainString, 'features'):
            logger.info('Features: ' + str(cfg.get('dqnpolicy_'+domainString, 'features')))
            self.policyfeatures = json.loads(cfg.get('dqnpolicy_'+domainString, 'features'))

        self.max_k = 5
        if cfg.has_option('dqnpolicy_'+domainString, 'max_k'):
            self.max_k = cfg.getint('dqnpolicy_'+domainString, 'max_k')

        self.learning_algorithm = 'drl'
        if cfg.has_option('dqnpolicy_'+domainString, 'learning_algorithm'):
            self.learning_algorithm = cfg.get('dqnpolicy_'+domainString, 'learning_algorithm')
            logger.info('Learning algorithm: ' + self.learning_algorithm)

        self.minibatch_size = 32
        if cfg.has_option('dqnpolicy_'+domainString, 'minibatch_size'):
            self.minibatch_size = cfg.getint('dqnpolicy_'+domainString, 'minibatch_size')

        self.capacity = 1000#max(self.minibatch_size, 2000)
        if cfg.has_option('dqnpolicy_'+domainString, 'capacity'):
            self.capacity = max(cfg.getint('dqnpolicy_'+domainString,'capacity'), 2000)

        self.replay_type = 'vanilla'
        if cfg.has_option('dqnpolicy_'+domainString, 'replay_type'):
            self.replay_type = cfg.get('dqnpolicy_'+domainString, 'replay_type')

        self.architecture = 'vanilla'
        if cfg.has_option('dqnpolicy_'+domainString, 'architecture'):
            self.architecture = cfg.get('dqnpolicy_'+domainString, 'architecture')

        self.q_update = 'single'
        if cfg.has_option('dqnpolicy_'+domainString, 'q_update'):
            self.q_update = cfg.get('dqnpolicy_'+domainString, 'q_update')

        self.h1_size = 130
        if cfg.has_option('dqnpolicy_'+domainString, 'h1_size'):
            self.h1_size = cfg.getint('dqnpolicy_'+domainString, 'h1_size')
        
        self.h2_size = 130
        if cfg.has_option('dqnpolicy_'+domainString, 'h2_size'):
            self.h2_size = cfg.getint('dqnpolicy_'+domainString, 'h2_size')

        """
        self.shuffle = False
        if cfg.has_option('dqnpolicy_'+domainString, 'experience_replay'):
            self.shuffle = cfg.getboolean('dqnpolicy_'+domainString, 'experience_replay')
        if not self.shuffle:
            # If we don't use experience replay, we don't need to maintain
            # sliding window of experiences with maximum capacity.
            # We only need to maintain the data of minibatch_size
            self.capacity = self.minibatch_size
        """

        self.episode_ave_max_q = []

        os.environ["CUDA_VISIBLE_DEVICES"]=""

        # init session
        self.sess = tf.Session()
        with tf.device("/cpu:0"):

            np.random.seed(self.randomseed)
            tf.set_random_seed(self.randomseed)

            # initialise an replay buffer
            if self.replay_type == 'vanilla':
                self.episodes[self.domainString] = ReplayBuffer(self.capacity, self.minibatch_size, self.randomseed)
            elif self.replay_type == 'prioritized':
                self.episodes[self.domainString] = ReplayPrioritised(self.capacity, self.minibatch_size, self.randomseed)
            #replay_buffer = ReplayBuffer(self.capacity, self.randomseed)
            #self.episodes = []
            self.samplecount = 0
            self.episodecount = 0

            # construct the models
            self.state_dim = self.n_in
            self.summaryaction = SummaryAction.SummaryAction(domainString)
            self.action_dim = len(self.summaryaction.action_names)
            action_bound = len(self.summaryaction.action_names)

            self.dqn = dqn.DeepQNetwork(self.sess, self.state_dim, self.action_dim, \
                self.critic_lr, self.tau, action_bound, self.architecture, self.h1_size, self.h2_size)

            # when all models are defined, init all variables
            init_op = tf.initialize_all_variables()
            self.sess.run(init_op)

            self.loadPolicy(self.in_policy_file)
            print('loaded replay size: ', self.episodes[self.domainString].size())

            self.dqn.update_target_network()

    # def record() has been handled...

    def act_on(self, beliefstate, hyps=None):
        if self.lastSystemAction is None and self.startwithhello:
            systemAct, nextaIdex = 'hello()', -1
        else:
            systemAct, nextaIdex = self.nextAction(beliefstate, hyps)
        self.lastSystemAction = systemAct
        self.summaryAct = nextaIdex
        self.prevbelief = beliefstate
        return systemAct

    def record(self, reward, domainInControl=None, weight=None, state=None, action=None):
        if domainInControl is None:
            domainInControl = self.domainString
        if self.episodes[domainInControl] is None:
            self.episodes[domainInControl] = Episode(dstring=domainInControl)
        if self.actToBeRecorded is None:
            #self.actToBeRecorded = self.lastSystemAction
            self.actToBeRecorded = self.summaryAct

        if state is None:
            state = self.prevbelief
        if action is None:
            action = self.actToBeRecorded

        cState, cAction = self.convertStateAction(state, action)

        # normalising total return to -1~1
        #reward /= 40.0
        reward /= 20.0
        """
        reward = float(reward+10.0)/40.0
        """

        if weight == None:
            if self.replay_type == 'vanilla':
                self.episodes[domainInControl].record(state=cState, \
                        state_ori=state, action=cAction, reward=reward)
            elif self.replay_type == 'prioritized':

                ##### calculate Q_s_t_a_t_ and gamma_Q_s_tplu1_maxa_ for PER ###
                ################################################################
                cur_cState = np.vstack([np.expand_dims(x, 0) for x in [cState]])        
                cur_cAction_one_hot = np.eye(self.action_dim, self.action_dim)[[cAction]]
                
                cur_action_q = self.dqn.predict(cur_cState, cur_cAction_one_hot)
                execMask = self.summaryaction.getExecutableMask(state, cAction)
               
                if self.q_update == 'single':
                    Qs = []
                    for idx, v in enumerate(execMask):
                        if v > -sys.maxsize:
                            Action_idx = np.eye(self.action_dim, self.action_dim)[[idx]]
                            Qidx = self.dqn.predict_target(cur_cState, Action_idx)
                            Qs.append(Qidx[0])
                            #Qs.append(Qidx[0])

                    Q_s_t_a_t_ = cur_action_q[0]
                    gamma_Q_s_tplu1_maxa_ = self.gamma * np.max(Qs)
                elif self.q_update == 'double':
                    Qs = []
                    for idx, v in enumerate(execMask):
                        if v > -sys.maxsize:
                            Action_idx = np.eye(self.action_dim, self.action_dim)[[idx]]
                            Qidx = self.dqn.predict(cur_cState, Action_idx)
                            Qs.append(Qidx[0])
                        else:
                            Qs.append(-sys.maxsize)

                    policyQ_argmax_a = np.argmax(Qs)
                    policyQ_argmax_a_one_hot = np.eye(self.action_dim, self.action_dim)[[policyQ_argmax_a]]
                    target_value_Q = self.dqn.predict_target(cur_cState, policyQ_argmax_a_one_hot)

                    Q_s_t_a_t_ = cur_action_q[0]
                    gamma_Q_s_tplu1_maxa_ = self.gamma * target_value_Q

                print('Q_s_t_a_t_', Q_s_t_a_t_)
                print('gamma_Q_s_tplu1_maxa_', gamma_Q_s_tplu1_maxa_)
                ################################################################

                # heuristically assign 0.0 to Q_s_t_a_t_ and Q_s_tplu1_maxa_, doesn't matter as it is not used
                #if self.samplecount >= self.capacity:
                if True:
                    self.episodes[domainInControl].record(state=cState, \
                        state_ori=state, action=cAction, reward=reward, \
                            Q_s_t_a_t_ = Q_s_t_a_t_, gamma_Q_s_tplu1_maxa_= gamma_Q_s_tplu1_maxa_, uniform=False)
                else:
                    self.episodes[domainInControl].record(state=cState, \
                        state_ori=state, action=cAction, reward=reward, \
                            Q_s_t_a_t_ = Q_s_t_a_t_, gamma_Q_s_tplu1_maxa_= gamma_Q_s_tplu1_maxa_, uniform=True)

        else:
            self.episodes[domainInControl].record(state=cState, state_ori=state, action=cAction, reward=reward, ma_weight=weight)

        self.actToBeRecorded = None
        self.samplecount += 1
        return

    def finalizeRecord(self, reward, domainInControl=None):
        if domainInControl is None:
            domainInControl = self.domainString
        if self.episodes[domainInControl] is None:
            logger.warning("record attempted to be finalized for domain where nothing has been recorded before")
            return

        #print 'Episode Avg_Max_Q', float(self.episode_ave_max_q)/float(self.episodes[domainInControl].size())
        print('Episode Avg_Max_Q', np.mean(self.episode_ave_max_q))

        print('saving statics')
        self.saveStats()
        print(self.stats)

        # normalising total return to -1~1
        #if reward == 0:
        #    reward = -20.0
        reward /= 20.0
        """
        if reward == 20.0:
            reward = 1.0
        else:
            reward = -0.5
        """
        #reward = float(reward+10.0)/40.0

        terminal_state, terminal_action = self.convertStateAction(TerminalState(), TerminalAction())

        if self.replay_type == 'vanilla':
            self.episodes[domainInControl].record(state=terminal_state, \
                    state_ori=TerminalState(), action=terminal_action, reward=reward, terminal=True)
        elif self.replay_type == 'prioritized':
            # heuristically assign 0.0 to Q_s_t_a_t_ and Q_s_tplu1_maxa_, doesn't matter as it is not used
                #if self.samplecount >= self.capacity:
                if True:
                    self.episodes[domainInControl].record(state=terminal_state, \
                        state_ori=TerminalState(), action=terminal_action, reward=reward, \
                            Q_s_t_a_t_ = 0.0, gamma_Q_s_tplu1_maxa_= 0.0, uniform=False, terminal=True)
                else:
                    self.episodes[domainInControl].record(state=terminal_state, \
                        state_ori=TerminalState(), action=terminal_action, reward=reward, \
                            Q_s_t_a_t_ = 0.0, gamma_Q_s_tplu1_maxa_= 0.0, uniform=True, terminal=True)
        return

    def convertStateAction(self, state, action):
        '''
        nnType = 'dnn'
        #nnType = 'rnn'
        # expand one dimension to match the batch size of 1 at axis 0
        if nnType == 'rnn':
            belief = np.expand_dims(belief,axis=0)
        '''
       
        if isinstance(state, TerminalState):
            return [0] * 260, action #260 #264
        else:
            flat_belief = flatten_belief(state)

            if flat_belief == self.prev_state_check:
                print('same state')
            else:
                print('diff state')
            self.prev_state_check = flat_belief

            return flat_belief, action

    def nextAction(self, beliefstate, hyps):
        '''
        select next action

        :param beliefstate: 
        :param hyps:
        :returns: (int) next summary action
        '''
        #beliefVec = flatten_belief(beliefstate, domainUtil)
        beliefVec = flatten_belief(beliefstate)

        execMask = self.summaryaction.getExecutableMask(beliefstate, self.lastSystemAction)
        #print sum([ 1 for i in execMask if i==0.0 ])
        if self.exploration_type == 'e-greedy':
            # epsilon greedy
            if self.is_training and utils.Settings.random.rand() < self.epsilon:
                admissible = [i for i, x in enumerate(execMask) if x == 0.0]
                random.shuffle(admissible)
                nextaIdex = admissible[0]
            else:
                admissible = []
                for idx, v in enumerate(execMask):
                    if v > -sys.maxsize:
                        Action_idx = np.eye(self.action_dim, self.action_dim)[[idx]]
                        Qidx = self.dqn.predict(np.reshape(beliefVec, (1, len(beliefVec))), Action_idx)
                        admissible.append(Qidx[0])
                    else:
                        admissible.append(-sys.maxsize)
                #action_Q = self.dqn.predict(np.reshape(beliefVec, (1, len(beliefVec))))# + (1. / (1. + i + j))
                #admissible = np.add(action_Q, np.array(execMask))
                logger.info('action Q...')
                print(admissible)
                nextaIdex = np.argmax(admissible)

                # add current max Q to self.episode_ave_max_q
                print('current maxQ', np.max(admissible))
                self.episode_ave_max_q.append(np.max(admissible))

        elif self.exploration_type == 'Boltzman':
            # randomly assign, not complete
            admissible = [i for i, x in enumerate(execMask) if x == 0.0]
            random.shuffle(admissible)
            nextaIdex = admissible[0]
      
        self.stats[nextaIdex] += 1
        summaryAct = self.summaryaction.action_names[nextaIdex]
        masterAct = self.summaryaction.Convert(beliefstate, summaryAct, self.lastSystemAction)
        return masterAct, nextaIdex

    def train(self):
        '''
        call this function when the episode ends
        '''

        if not self.is_training:
            logger.info("Not in training mode")
            return
        else:
            logger.info("Update dqn policy parameters.")

        self.episodecount += 1
        logger.info("Sample Num so far: %s" %(self.samplecount))
        logger.info("Episode Num so far: %s" %(self.episodecount))
        #if True:
        if self.samplecount >= self.minibatch_size * 3 and self.episodecount % 4 == 0:
        #if self.samplecount >= self.capacity and self.episodecount % 5 == 0:
        #if self.samplecount > self.minibatch_size:
        #if self.samplecount > self.capacity:
            logger.info('start traninig...')


            #################################################
            #################################################
            #   update TD error for all experience in PER   #
            #################################################
            #################################################
            """
            #s_batch, s_ori_batch, a_batch, r_batch, s2_batch, s2_ori_batch, t_batch, idx_batch, _ = \
            #    self.episodes[self.domainString].all_batch()
            experience, idx_batch = self.episodes[self.domainString].all_batch()
                #self.episodes[self.domainString].sample_batch_vanilla_PER()
           
            #s_batch = np.vstack([np.expand_dims(x, 0) for x in s_batch])
            #s2_batch = np.vstack([np.expand_dims(x, 0) for x in s2_batch])

            # self.s_prev, self.s_ori_prev, self.a_prev, self.r_prev, state, state_ori, termina
    
            for k in xrange(len(idx_batch)):
                Q_bootstrap_label = 0
                if experience[k][-1]: # terminal
                    Q_bootstrap_label = experience[k][3] # reward
                else:
                    execMask = self.summaryaction.getExecutableMask(experience[k][-2], experience[k][2]) # s_ori, a
                    if self.q_update == 'single':
                        admissible = []
                        for idx, v in enumerate(execMask):
                            if v > -sys.maxint:
                                Action_idx = np.eye(self.action_dim, self.action_dim)[[idx]]
                                s2_idx = np.vstack([ np.expand_dims(x, 0) for x in [experience[k][0]] ]) # s
                                Qidx = self.dqn.predict_target(s2_idx, Action_idx)
                                admissible.append(Qidx[0])
                        Q_bootstrap_label = experience[k][3] + self.gamma * np.max(admissible) # reward
                    elif self.q_update == 'double':
                        Qs = []
                        for idx, v in enumerate(execMask):
                            if v > -sys.maxint:
                                Action_idx = np.eye(self.action_dim, self.action_dim)[[idx]]
                                s2_idx = np.vstack([ np.expand_dims(x, 0) for x in [experience[k][-3]] ])
                                Qidx = self.dqn.predict(s2_idx, Action_idx)
                                Qs.append(Qidx[0])
                            else:
                                Qs.append(-sys.maxint)

                        policyQ_argmax_a = np.argmax(Qs)
                        policyQ_argmax_a_one_hot = np.eye(self.action_dim, self.action_dim)[[policyQ_argmax_a]]
                        s2_idx = np.vstack([ np.expand_dims(x, 0) for x in [experience[k][-3]] ])
                        target_value_Q = self.dqn.predict_target(s2_idx, policyQ_argmax_a_one_hot)

                        Q_bootstrap_label = experience[k][3] + self.gamma * target_value_Q

                if self.replay_type == 'prioritized':
                    # update the sum-tree
                    # update the TD error of the samples in the minibatch
                    current_a = np.eye(self.action_dim, self.action_dim)[[experience[k][2]]]
                    current_s = np.vstack([ np.expand_dims(x, 0) for x in [experience[k][0]] ])
                    currentQ_s_a_ = self.dqn.predict(current_s, current_a)
                    currentQ_s_a_ = currentQ_s_a_[0]
                    error = abs(currentQ_s_a_ - Q_bootstrap_label)
                    self.episodes[self.domainString].update(idx_batch[k], error)
 
            """

            s_batch, s_ori_batch, a_batch, r_batch, s2_batch, s2_ori_batch, t_batch, idx_batch, _ = \
                self.episodes[self.domainString].sample_batch()
                #self.episodes[self.domainString].sample_batch_vanilla_PER()
           
            #s_batch = np.vstack([np.expand_dims(x, 0) for x in s_batch])
            #s2_batch = np.vstack([np.expand_dims(x, 0) for x in s2_batch])

            y_i = []
            for k in range(min(self.minibatch_size, self.episodes[self.domainString].size())):
                Q_bootstrap_label = 0
                if t_batch[k]:
                    Q_bootstrap_label = r_batch[k]
                else:
                    execMask = self.summaryaction.getExecutableMask(s2_ori_batch[k], a_batch[k])
                    if self.q_update == 'single':
                        admissible = []
                        for idx, v in enumerate(execMask):
                            if v > -sys.maxsize:
                                Action_idx = np.eye(self.action_dim, self.action_dim)[[idx]]
                                s2_idx = np.vstack([ np.expand_dims(x, 0) for x in [s2_batch[k]] ])
                                Qidx = self.dqn.predict_target(s2_idx, Action_idx)
                                admissible.append(Qidx[0])
                        Q_bootstrap_label = r_batch[k] + self.gamma * np.max(admissible)
                    elif self.q_update == 'double':
                        Qs = []
                        for idx, v in enumerate(execMask):
                            if v > -sys.maxsize:
                                Action_idx = np.eye(self.action_dim, self.action_dim)[[idx]]
                                s2_idx = np.vstack([ np.expand_dims(x, 0) for x in [s2_batch[k]] ])
                                Qidx = self.dqn.predict(s2_idx, Action_idx)
                                Qs.append(Qidx[0])
                            else:
                                Qs.append(-sys.maxsize)

                        policyQ_argmax_a = np.argmax(Qs)
                        policyQ_argmax_a_one_hot = np.eye(self.action_dim, self.action_dim)[[policyQ_argmax_a]]
                        s2_idx = np.vstack([ np.expand_dims(x, 0) for x in [s2_batch[k]] ])
                        target_value_Q = self.dqn.predict_target(s2_idx, policyQ_argmax_a_one_hot)

                        Q_bootstrap_label = r_batch[k] + self.gamma * target_value_Q
                y_i.append(Q_bootstrap_label)

                if self.replay_type == 'prioritized':
                    # update the sum-tree
                    # update the TD error of the samples in the minibatch
                    current_a = np.eye(self.action_dim, self.action_dim)[[a_batch[k]]]
                    current_s = np.vstack([ np.expand_dims(x, 0) for x in [s_batch[k]] ])
                    currentQ_s_a_ = self.dqn.predict(current_s, current_a)
                    currentQ_s_a_ = currentQ_s_a_[0]
                    error = abs(currentQ_s_a_ - Q_bootstrap_label)
                    self.episodes[self.domainString].update(idx_batch[k], error)
                                         
            # change index-based a_batch to one-hot-based a_batch
            a_batch_one_hot = np.eye(self.action_dim, self.action_dim)[a_batch]

            # Update the critic given the targets
            reshaped_yi = np.vstack([np.expand_dims(x, 0) for x in y_i])

            s_batch_expand = np.vstack([np.expand_dims(x, 0) for x in s_batch])
            """
            print s_batch_expand
            print 'a_batch', a_batch
            print a_batch_one_hot
            print len(a_batch)
            print len(y_i)
            """
            #reshaped_yi = np.reshape(y_i, (min(self.minibatch_size, self.episodes[self.domainString].size()), 1))
            predicted_q_value, _, currentLoss = self.dqn.train(s_batch, a_batch_one_hot, reshaped_yi)
            #predicted_q_value, _, currentLoss = self.dqn.train(s_batch_expand, a_batch_one_hot, reshaped_yi)
            
            print('y_i')
            print(y_i)
            print('currentLoss', currentLoss)
            print('predict Q')
            print(predicted_q_value)

            if self.episodecount % 1 == 0:
            #if self.episodecount % 50 == 0:
            # Update target networks
                self.dqn.update_target_network()

            self.savePolicyInc()  # self.out_policy_file)

    def savePolicy(self, FORCE_SAVE=False):
        """
        Does not use this, cause it will be called from agent after every episode.
        we want to save the policy only periodically.
        """
        pass

    def savePolicyInc(self, FORCE_SAVE=False):
        """
        save model and replay buffer
        """
        #save_path = self.saver.save(self.sess, self.out_policy_file+'.ckpt')
        self.dqn.save_network(self.out_policy_file+'.dqn.ckpt')

        f = open(self.out_policy_file+'.episode', 'wb')
        for obj in [self.samplecount, self.episodes[self.domainString]]:
            pickle.dump(obj, f, protocol=pickle.HIGHEST_PROTOCOL)
        f.close()
        #logger.info("Saving model to %s and replay buffer..." % save_path)

    def saveStats(self, FORCE_SAVE=False):
        f = open(self.out_policy_file + '.stats', 'wb')
        pickle.dump(self.stats, f, protocol=pickle.HIGHEST_PROTOCOL)
        f.close()

    def loadPolicy(self, filename):
        """
        load model and replay buffer
        """
        # load models
        self.dqn.load_network(filename+'.dqn.ckpt')
        
        # load replay buffer
        try:
            print('laod from: ', filename)
            f = open(filename+'.episode', 'rb')
            loaded_objects = []
            for i in range(2): # load nn params and collected data
                loaded_objects.append(pickle.load(f))
            self.samplecount = int(loaded_objects[0])
            self.episodes[self.domainString] = copy.deepcopy(loaded_objects[1])
            logger.info("Loading both model from %s and replay buffer..." % filename)
            f.close()
        except:
            logger.info("Loading only models...")

    def restart(self):
        self.summaryAct = None          
        self.lastSystemAction = None
        self.prevbelief = None
        self.actToBeRecorded = None
        self.epsilon = self.epsilon_start - (self.epsilon_start - self.epsilon_end) * float(self.episodeNum+self.episodecount) / float(self.maxiter)
        print('current eps', self.epsilon)
        #self.episodes = dict.fromkeys(OntologyUtils.available_domains, None)
        #self.episodes[self.domainString] = ReplayBuffer(self.capacity, self.randomseed)
        self.episode_ave_max_q = []

#END OF FILE
