###############################################################################
# PyDial: Multi-domain Statistical Spoken Dialogue System Software
###############################################################################
#
# Copyright 2015 - 2019
# Cambridge University Engineering Department Dialogue Systems Group
#
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

'''
ENACPolicy.py - Advantage Actor-Critic policy
==================================================

Copyright CUED Dialogue Systems Group 2015 - 2017

.. seealso:: CUED Imports/Dependencies: 

    import :class:`Policy`
    import :class:`utils.ContextLogger`

.. warning::
        Documentation not done.


************************

'''

import copy
import os
import json
import numpy as np
import scipy
import scipy.signal
import pickle as pickle
import random
import utils
from utils.Settings import config as cfg
from utils import ContextLogger, DiaAct

import ontology.FlatOntologyManager as FlatOnt
#from theano_dialogue.util.tool import *

import tensorflow as tf
from policy.DRL.replay_buffer_episode_enac import ReplayBufferEpisode
from policy.DRL.replay_prioritised_episode import ReplayPrioritisedEpisode
import policy.DRL.utils as drlutils
import policy.DRL.enac as enac
import policy.Policy
from policy.ENACPolicy import ENACPolicy
import policy.SummaryAction
from policy.Policy import TerminalAction, TerminalState
from policy.feudalRL.DIP_parametrisation import DIP_state, padded_state
from policy.feudalRL.feudalUtils import get_feudal_masks

logger = utils.ContextLogger.getLogger('')


# Discounting function used to calculate discounted returns.
def discount(x, gamma):
    return scipy.signal.lfilter([1], [1, -gamma], x[::-1], axis=0)[::-1]

class FeudalENACPolicy(ENACPolicy):
    '''Derived from :class:`Policy`
    '''
    def __init__(self, in_policy_file, out_policy_file, domainString='CamRestaurants', is_training=False, action_names=None, slot=None):
        super(FeudalENACPolicy, self).__init__(in_policy_file, out_policy_file, domainString=domainString, is_training=is_training)

        tf.reset_default_graph()

        self.in_policy_file = in_policy_file
        self.out_policy_file = out_policy_file
        self.is_training = is_training
        self.accum_belief = []
        self.prev_state_check = None

        self.domainString = domainString
        self.domainUtil = FlatOnt.FlatDomainOntology(self.domainString)

        self.features = 'dip'
        self.sd_enc_size = 80
        self.si_enc_size = 40
        self.dropout_rate = 0.
        if cfg.has_option('feudalpolicy', 'features'):
            self.features = cfg.get('feudalpolicy', 'features')
        if cfg.has_option('feudalpolicy', 'sd_enc_size'):
            self.sd_enc_size = cfg.getint('feudalpolicy', 'sd_enc_size')
        if cfg.has_option('feudalpolicy', 'si_enc_size'):
            self.si_enc_size = cfg.getint('feudalpolicy', 'si_enc_size')
        if cfg.has_option('dqnpolicy', 'dropout_rate') and self.is_training:
            self.dropout_rate = cfg.getfloat('feudalpolicy', 'dropout_rate')
        if cfg.has_option('dqnpolicy', 'dropout_rate') and self.is_training:
            self.dropout_rate = cfg.getfloat('feudalpolicy', 'dropout_rate')


        # init session
        self.sess = tf.Session()
        with tf.device("/cpu:0"):

            np.random.seed(self.randomseed)
            tf.set_random_seed(self.randomseed)

            # initialise an replay buffer
            if self.replay_type == 'vanilla':
                self.episodes[self.domainString] = ReplayBufferEpisode(self.capacity, self.minibatch_size, self.randomseed)
            elif self.replay_type == 'prioritized':
                self.episodes[self.domainString] = ReplayPrioritisedEpisode(self.capacity, self.minibatch_size, self.randomseed)
            #replay_buffer = ReplayBuffer(self.capacity, self.randomseed)
            #self.episodes = []
            self.samplecount = 0
            self.episodecount = 0

            # construct the models
            self.state_dim = 89  # current DIP state dim
            self.summaryaction = policy.SummaryAction.SummaryAction(domainString)
            self.action_names = action_names
            self.action_dim = len(self.action_names)
            action_bound = len(self.action_names)
            self.stats = [0 for _ in range(self.action_dim)]

            if self.features == 'dip':
                self.enac = enac.ENACNetwork(self.sess, self.state_dim, self.action_dim, self.critic_lr, self.tau,
                                         action_bound, self.architecture, self.h1_size, self.h2_size, self.is_training)
            elif self.features == 'learned' or self.features == 'rnn':
                si_state_dim = 72
                if self.domainString == 'CamRestaurants':
                    sd_state_dim = 94
                elif self.domainString == 'SFRestaurants':
                    sd_state_dim = 158
                elif self.domainString == 'Laptops11':
                    sd_state_dim = 13
                else:
                    logger.error(
                        'Domain {} not implemented in feudal-DQN yet')  # just find out the size of sd_state_dim for the new domain
                if self.features == 'rnn':
                    arch = 'rnn'
                else:
                    arch = 'vanilla'
                self.state_dim = si_state_dim + sd_state_dim
                self.enac = enac.ENACNetwork(self.sess, self.state_dim, self.action_dim, self.critic_lr, self.tau,
                                             action_bound, self.architecture, self.h1_size, self.h2_size, self.is_training)
            else:
                logger.error('features "{}" not implemented'.format(self.features))

            # when all models are defined, init all variables
            init_op = tf.global_variables_initializer()
            self.sess.run(init_op)

            self.loadPolicy(self.in_policy_file)
            print('loaded replay size: ', self.episodes[self.domainString].size())


    def convertStateAction(self, state, action):
        '''

        '''
        if isinstance(state, TerminalState):
            return [0] * 89, action

        else:
            if self.features == 'learned' or self.features == 'rnn':
                dip_state = padded_state(state.domainStates[state.currentdomain], self.domainString)
            else:
                dip_state = DIP_state(state.domainStates[state.currentdomain], self.domainString)
            action_name = self.actions.action_names[action]
            act_slot = 'general'
            for slot in dip_state.slots:
                if slot in action_name:
                    act_slot = slot
            flat_belief = dip_state.get_beliefStateVec(act_slot)
            self.prev_state_check = flat_belief

            return flat_belief, action

    def record(self, reward, domainInControl=None, weight=None, state=None, action=None):
        if domainInControl is None:
            domainInControl = self.domainString
        if self.actToBeRecorded is None:
            #self.actToBeRecorded = self.lastSystemAction
            self.actToBeRecorded = self.summaryAct

        if state is None:
            state = self.prevbelief
        if action is None:
            action = self.actToBeRecorded

        cState, cAction = state, action

        # normalising total return to -1~1
        reward /= 20.0

        #value = self.a2c.predict_value([cState])
        value = np.array([[0.0]])
        policy_mu = self.mu_prob

        if weight == None:
            if self.replay_type == 'vanilla':
                self.episodes[domainInControl].record(state=cState, \
                        state_ori=state, action=cAction, reward=reward, value=value[0][0], distribution=policy_mu)
            elif self.replay_type == 'prioritized':
                self.episodes[domainInControl].record(state=cState, \
                        state_ori=state, action=cAction, reward=reward, value=value[0][0], distribution=policy_mu)
        else:
            self.episodes[domainInControl].record(state=cState, state_ori=state, action=cAction, reward=reward, ma_weight=weight)

        self.actToBeRecorded = None
        self.samplecount += 1
        return

    def nextAction(self, beliefstate):
        '''
        select next action

        :param beliefstate:
        :returns: (int) next summary action
        '''

        if self.exploration_type == 'e-greedy':

            # epsilon greedy
            if self.is_training and utils.Settings.random.rand() < self.epsilon:
                action_prob = np.random.rand(len(self.action_names))

                # Importance sampling (should be turned off)
                #if nextaIdex == greedyNextaIdex:
                #    self.mu_prob = self.epsilon / float(self.action_dim) + 1 - self.epsilon
                #else:
                #    self.mu_prob = self.epsilon / float(self.action_dim)
            else:
                action_prob = self.enac.predict_policy(np.reshape(beliefstate, (1, len(beliefstate))))

                # add current max Q to self.episode_ave_max_q
                #print 'current maxQ', np.max(admissible)
                #self.episode_ave_max_q.append(np.max(admissible))
                
                # Importance sampling
                #self.mu_prob = self.epsilon / float(self.action_dim) + 1 - self.epsilon

        return action_prob

    def train(self):
        '''
        call this function when the episode ends
        '''

        if not self.is_training:
            logger.info("Not in training mode")
            return
        else:
            logger.info("Update enac policy parameters.")

        self.episodecount += 1
        logger.info("Sample Num so far: %s" %(self.samplecount))
        logger.info("Episode Num so far: %s" %(self.episodecount))

        if self.samplecount >= self.minibatch_size and self.episodecount % self.training_frequency == 0:
            logger.info('start training...')

            s_batch, s_ori_batch, a_batch, r_batch, s2_batch, s2_ori_batch, t_batch, idx_batch, v_batch, mu_policy = \
                self.episodes[self.domainString].sample_batch()

            discounted_return_batch = []
        

            def weightsImportanceSampling(mu_policy, r_batch):
                mu_policy = np.asarray(mu_policy)
                mu_cum = []
                lenghts = []  # to properly divide on dialogues pi_policy later on
                for mu in mu_policy:
                    lenghts.append(len(mu))
                    mu = np.asarray(mu).astype(np.longdouble)
                    mu_cum.append(np.cumprod(mu[::-1])[::-1])  # going forward with cumulative product
                # mu_cum = np.concatenate(np.array(mu_cum), axis=0).tolist()
                mu_policy = np.concatenate(np.array(mu_policy), axis=0).tolist()  # concatenate all behavioral probs
                lengths = np.cumsum(lenghts)  # time steps for ends of dialogues
                lengths = np.concatenate((np.array([0]), lengths), axis=0)  # add first dialogue

                if self.importance_sampling == 'max':
                    pass
                elif self.importance_sampling == "soft":
                    # get the probabilities of actions taken from the batch
                    pi_policy = self.enac.getPolicy(np.concatenate(np.array(s_batch), axis=0).tolist())[0]  # policy given s_t
                    columns = np.asarray([np.concatenate(a_batch, axis=0).tolist()]).astype(int)  # actions taken at s_t
                    rows = np.asarray([ii for ii in range(len(pi_policy))])
                    pi_policy = pi_policy[rows, columns][0].astype(np.longdouble)  # getting probabilities for current policy

                #####################################
                # Weights for importance sampling
                # it goes through each dialogue and computes in reverse order cumulative prod:
                # rho_n = pi_n / mu_n
                # ...
                # rho_1 = pi_1 / mu_1 *  ... * pi_n / mu_n
                # using dialogue and weight_cum lists
                #####################################

                rho_forward = []  # rho_forward from eq. 3.3 (the first one)
                rho_whole = []  # product across the whole dialogue from eq. 3.3 (the second one)
                #pi_cum2 = []  # stats to compare
                #mu_cum2 = []  # stats to compare
                #pi_cum = []  # stats to compare

                # Precup version
                r_vector = np.concatenate(np.array(r_batch), axis=0).tolist()
                r_weighted = []

                for ii in range(len(lengths) - 1):  # over dialogues
                    weight_cum = 1.
                    dialogue = []

                    for pi, mu in zip(pi_policy[lengths[ii]:lengths[ii + 1]], mu_policy[lengths[ii]:lengths[ii + 1]]):
                        weight_cum *= pi / mu
                        dialogue.append(weight_cum)

                    dialogue = np.array(dialogue)
                    dialogue = np.clip(dialogue, 0.5, 1)  # clipping the weights
                    dialogue = dialogue.tolist()

                    rho_forward.extend(dialogue)
                    #rho_whole.append(dialogue[-1])
                    rho_whole.extend(np.ones(len(dialogue)) * dialogue[-1])
                    r_weighted.extend(r_vector[lengths[ii]: lengths[ii + 1]] * np.asarray(dialogue))

                # go back to original form:
                ind = 0
                r_new = copy.deepcopy(r_batch)
                for id, batch in enumerate(r_new):
                    for id2, _ in enumerate(batch):
                        r_new[id][id2] = r_weighted[ind]
                        ind += 1

                # ONE STEP WEIGHTS
                weights = np.asarray(pi_policy) / np.asarray(mu_policy)
                weights = np.clip(weights, 0.5, 1)  # clipping the weights

                return weights, rho_forward, rho_whole, r_new

            weights, rho_forward, rho_whole, r_new = weightsImportanceSampling(mu_policy, r_batch)

            weights = np.nan_to_num(weights)
            rho_forward = np.nan_to_num(rho_forward)
            rho_whole = np.nan_to_num(rho_whole)
            """
            print 'w',weights
            print 'rho_for',rho_forward
            print 'rho_who',rho_whole
            """

            def calculate_discountR(r_episode, idx):
                #########################################################################
                # Here we take the rewards and values from the rollouts, and use them to
                # generate the advantage and discounted returns.
                # The advantage function uses "Generalized Advantage Estimation"
                bootstrap_value = 0.0
                # r_episode rescale by rhos?
                self.r_episode_plus = np.asarray(r_episode[idx:] + [bootstrap_value])
                if self.importance_sampling:
                    self.r_episode_plus = self.r_episode_plus
                else:
                    self.r_episode_plus = self.r_episode_plus/rho_forward[idx]
                discounted_r_episode = discount(self.r_episode_plus, self.gamma)[:-1]
                #########################################################################
                return discounted_r_episode[0]

            if self.replay_type == 'prioritized':
                for item_r, item_v, item_idx in zip(r_new, v_batch, idx_batch):
                    rlist = []
                    for idx in range(len(item_r)):
                        r = calculate_discountR(item_r, idx)
                        rlist.append(r)

                    discounted_return_batch.append(rlist[-1])
            else:
                for item_r, item_v in zip(r_new, v_batch):
                    rlist = []
                    for idx in range(len(item_r)):
                        r = calculate_discountR(item_r, idx)
                        rlist.append(r)

                    discounted_return_batch.append(rlist[-1]) 

            batch_size = len(s_batch)

            if self.importance_sampling:
                discounted_return_batch = np.clip(discounted_return_batch, -1, 1)

            # get gradient info and create matrix
            gradient_matrix = []
            for item_s, item_a in zip(s_batch, a_batch):
                item_a_one_hot = np.eye(self.action_dim)[item_a]
                policy_gradient = self.enac.get_policy_gradient(item_s, item_a_one_hot)
                policy_gradient = [(policy_gradient_idv.flatten()).tolist() for policy_gradient_idv in policy_gradient]
                policy_gradient_flatten = np.hstack(policy_gradient)
                policy_gradient_flatten = np.append(policy_gradient_flatten, [1.0])
                gradient_matrix.append(policy_gradient_flatten.tolist())
            
            gradient_matrix = np.matrix(gradient_matrix)
            return_matrix = np.matrix(discounted_return_batch)

            logger.info("Updating eNAC policy parameters, before calculate eNac matrix")
            try: 
                natural_gradient = np.dot(np.linalg.pinv(gradient_matrix), return_matrix.T)
                # convert a matrix to list-like array
                natural_gradient = np.array(natural_gradient.flatten()).ravel()
                natural_gradient = natural_gradient[:-1] # discard the last element
            except:
                natural_gradient = self.natural_gradient_prev 
                print('SVD problem')

            logger.info("Updating eNAC policy parameters, after calculate eNac matrix")

            self.natural_gradient_prev = natural_gradient

            all_params = self.enac.get_params()

            cnt = 0
            modelW = []
            modelB = []
            for variable in all_params:
                       
                shape = variable.shape
                # weight matrix
                if np.array(variable).ndim == 1:
                    until = np.array(variable).shape[0]
                    subNG = np.reshape(natural_gradient[cnt:cnt+until],shape)
                    cnt += until
                    modelB.append(subNG)
                # bias vector
                elif np.array(variable).ndim == 2:
                    until = np.array(variable).shape[0]*np.array(variable).shape[1]
                    subNG = np.reshape(natural_gradient[cnt:cnt+until],shape)
                    cnt += until
                    modelW.append(subNG)

            a_batch_one_hot = np.eye(self.action_dim)[np.concatenate(a_batch, axis=0).tolist()]

            policy_loss, entropy, all_loss, optimise = self.enac.train( \
                    np.concatenate(np.array(s_batch), axis=0).tolist(), a_batch_one_hot, \
                    modelW[0], modelB[0], modelW[1], modelB[1], modelW[2], modelB[2] \
            )

            norm_p_l, ent, norm_all_l = \
                    policy_loss/float(batch_size), \
                    entropy/float(batch_size), all_loss/float(batch_size)

        self.savePolicyInc()  # self.out_policy_file)

    def savePolicy(self, FORCE_SAVE=False):
        """
        Does not use this, cause it will be called from agent after every episode.
        we want to save the policy only periodically.
        """
        pass

    def savePolicyInc(self, FORCE_SAVE=False):
        """
        save model and replay buffer
        """
        if self.episodecount % self.save_step == 0:
            self.enac.save_network(self.out_policy_file+'.enac.ckpt')

            f = open(self.out_policy_file+'.episode', 'wb')
            for obj in [self.samplecount, self.episodes[self.domainString]]:
                pickle.dump(obj, f, protocol=pickle.HIGHEST_PROTOCOL)
            f.close()
            #logger.info("Saving model to %s and replay buffer..." % save_path)

    def saveStats(self, FORCE_SAVE=False):
        f = open(self.out_policy_file + '.stats', 'wb')
        pickle.dump(self.stats, f, protocol=pickle.HIGHEST_PROTOCOL)
        f.close()

    def loadPolicy(self, filename):
        """
        load model and replay buffer
        """
        # load models
        self.enac.load_network(filename+'.enac.ckpt')
        
        # load replay buffer
        try:
            print('load from: ', filename)
            f = open(filename+'.episode', 'rb')
            loaded_objects = []
            for i in range(2): # load nn params and collected data
                loaded_objects.append(pickle.load(f))
            self.samplecount = int(loaded_objects[0])
            self.episodes[self.domainString] = copy.deepcopy(loaded_objects[1])
            logger.info("Loading both model from %s and replay buffer..." % filename)
            f.close()
        except:
            logger.info("Loading only models...")

    def restart(self):
        self.summaryAct = None          
        self.lastSystemAction = None
        self.prevbelief = None
        self.actToBeRecorded = None
        self.epsilon = self.epsilon_start - (self.epsilon_start - self.epsilon_end) * float(self.episodeNum+self.episodecount) / float(self.maxiter)
        self.episode_ave_max_q = []

#END OF FILE
